import itertools
import time

import numpy
import pytest

from helpers import *

from reikna.helpers import product
from reikna.fft import FFTShift
import reikna.cluda.dtypes as dtypes
from reikna.transformations import mul_param


def pytest_generate_tests(metafunc):

    errors_shapes_and_axes = [
        ((10,), (0,)),
        ((11,), (0,)),
        ((9000,), (0,)),
        ((9001,), (0,)),
        ((128, 60), (0, 1)),
        ((127, 60), (0, 1)),
        ((127, 61), (0, 1)),
        ((100, 80, 60), (0, 1, 2)),
        ((101, 80, 61), (0, 1, 2)),
        ((101, 80, 61), (0, 2)),
        ((20, 31, 80, 61), (0, 2)),
        ]

    perf_shapes = [
        (2**4,), # 1D, small size
        (2**18,), # 1D, large size

        (2**4, 2**4), # 2D, small size
        (2**9, 2**9), # 2D, large size
        ]
    perf_even_shapes_and_axes = []
    perf_odd_shapes_and_axes = []

    mem_limit = 2**22

    for contigous in (True, False):
        for shape in perf_shapes:
            batch = mem_limit // product(shape)
            if contigous:
                full_shape = (batch,) + shape
                axes = tuple(range(1, len(shape) + 1))
            else:
                full_shape = shape + (batch,)
                axes = tuple(range(0, len(shape)))

            perf_even_shapes_and_axes.append((full_shape, axes))

            full_shape = list(full_shape)
            for axis in axes:
                full_shape[axis] -= 1
            perf_odd_shapes_and_axes.append((tuple(full_shape), axes))

    idgen = lambda pair: str(pair[0]) + '_over_' + str(pair[1])

    if 'errors_shape_and_axes' in metafunc.funcargnames:
        metafunc.parametrize(
            'errors_shape_and_axes', errors_shapes_and_axes,
            ids=list(map(idgen, errors_shapes_and_axes)))

    elif 'perf_even_shape_and_axes' in metafunc.funcargnames:
        metafunc.parametrize(
            'perf_even_shape_and_axes', perf_even_shapes_and_axes,
            ids=list(map(idgen, perf_even_shapes_and_axes)))

    elif 'perf_odd_shape_and_axes' in metafunc.funcargnames:
        metafunc.parametrize(
            'perf_odd_shape_and_axes', perf_odd_shapes_and_axes,
            ids=list(map(idgen, perf_odd_shapes_and_axes)))


def check_errors(thr, shape_and_axes):

    dtype = numpy.int32

    shape, axes = shape_and_axes

    data = numpy.arange(product(shape)).reshape(shape).astype(dtype)

    shift = FFTShift(data, axes=axes)
    shiftc = shift.compile(thr)

    #print(shiftc._kernel_calls[0]._kernel._program.source)
    #print(shiftc._kernel_calls[0]._kernel.global_size)

    data_dev = thr.to_device(data)
    shiftc(data_dev, data_dev)
    res_ref = numpy.fft.fftshift(data, axes=axes)
    #print(data)
    #print(res_ref)
    #print(data_dev.get())
    assert diff_is_negligible(data_dev.get(), res_ref)


def test_errors(thr, errors_shape_and_axes):
    check_errors(thr, errors_shape_and_axes)


def test_trivial(some_thr):
    """
    Checks that even if the axes set is trivial (product of lengths == 1),
    the transformations are still attached and executed.
    """
    dtype = numpy.complex64
    shape = (128, 1, 1, 128)
    axes = (1, 2)
    param = 4

    data = get_test_array(shape, dtype)
    data_dev = some_thr.to_device(data)
    res_dev = some_thr.empty_like(data_dev)

    shift = FFTShift(data, axes=axes)
    scale = mul_param(data_dev, numpy.int32)
    shift.parameter.input.connect(scale, scale.output, input_prime=scale.input, param=scale.param)

    shiftc = shift.compile(some_thr)
    shiftc(res_dev, data_dev, param)
    assert diff_is_negligible(res_dev.get(), data * param)


def test_unordered_axes(some_thr):
    check_errors(some_thr, ((40, 50, 60), (2, 0)))


def test_no_axes(some_thr):
    check_errors(some_thr, ((40, 50, 60), None))


def check_performance(thr_and_double, shape_and_axes):

    thr, double = thr_and_double

    dtype = numpy.complex128 if double else numpy.complex64
    dtype = dtypes.normalize_type(dtype)

    shape, axes = shape_and_axes

    data = numpy.arange(product(shape)).reshape(shape).astype(dtype)

    shift = FFTShift(data, axes=axes)
    shiftc = shift.compile(thr)

    data_dev = thr.to_device(data)
    res_dev = thr.empty_like(data)

    attempts = 10
    times = []
    for i in range(attempts):
        t1 = time.time()
        shiftc(res_dev, data_dev)
        thr.synchronize()
        times.append(time.time() - t1)

    res_ref = numpy.fft.fftshift(data, axes=axes)
    assert diff_is_negligible(res_dev.get(), res_ref)

    return min(times), product(shape) * dtype.itemsize


@pytest.mark.perf
@pytest.mark.returns('GB/s')
def test_even_performance(thr_and_double, perf_even_shape_and_axes):
    return check_performance(thr_and_double, perf_even_shape_and_axes)


@pytest.mark.perf
@pytest.mark.returns('GB/s')
def test_odd_performance(thr_and_double, perf_odd_shape_and_axes):
    return check_performance(thr_and_double, perf_odd_shape_and_axes)



"""
This example illustrates how to:
- attach a transformation to an FFT computation object that will make it
  operate on real-valued inputs.
"""

import numpy
from reikna.cluda import dtypes, any_api
from reikna.fft import FFT
from reikna.core import Annotation, Type, Transformation, Parameter


# Pick the first available GPGPU API and make a Thread on it.
api = any_api()
thr = api.Thread.create()


# A transformation that transforms a real array to a complex one
# by adding a zero imaginary part
def get_complex_trf(arr):
    complex_dtype = dtypes.complex_for(arr.dtype)
    return Transformation(
        [Parameter('output', Annotation(Type(complex_dtype, arr.shape), 'o')),
        Parameter('input', Annotation(arr, 'i'))],
        """
        ${output.store_same}(
            COMPLEX_CTR(${output.ctype})(
                ${input.load_same},
                0));
        """)


arr = numpy.random.normal(size=3000).astype(numpy.float32)

trf = get_complex_trf(arr)


# Create the FFT computation and attach the transformation above to its input.
fft = FFT(trf.output) # (A shortcut: using the array type saved in the transformation)
fft.parameter.input.connect(trf, trf.output, new_input=trf.input)
cfft = fft.compile(thr)


# Run the computation
arr_dev = thr.to_device(arr)
res_dev = thr.array(arr.shape, numpy.complex64)
cfft(res_dev, arr_dev)
result = res_dev.get()

reference = numpy.fft.fft(arr)

assert numpy.linalg.norm(result - reference) / numpy.linalg.norm(reference) < 1e-6



import numpy
import pytest

from helpers import *
from reikna.linalg import EntrywiseNorm


def reference_norm(arr, order=1, axes=None):
    if axes is None:
        axes = list(range(len(arr.shape)))

    arr = numpy.abs(arr)
    out_dtype = arr.dtype

    arr = arr ** order
    for axis in reversed(sorted(axes)):
        arr = arr.sum(axis)

    # explicit cast to preven numpy promoting arr to float64 from float32 if it is 0-dimensional
    res = arr ** numpy.cast[out_dtype](1. / order)
    return res


def check_norm(thr, shape, dtype, order, axes):
    a = get_test_array(shape, dtype)
    a_dev = thr.to_device(a)

    norm = EntrywiseNorm(a_dev, order=order, axes=axes)

    b_dev = thr.empty_like(norm.parameter.output)
    b_ref = reference_norm(a, order=order, axes=axes)

    normc = norm.compile(thr)
    normc(b_dev, a_dev)

    assert diff_is_negligible(b_dev.get(), b_ref)


@pytest.mark.parametrize('dtype', [numpy.float32, numpy.complex64], ids=['float32', 'complex64'])
@pytest.mark.parametrize('order', [0.5, 1, 2])
def test_all_axes(thr, dtype, order):
    check_norm(thr, 100, dtype, order, None)


axes_vals = [None, (0,), (2,), (1, 2)]
axes_ids = [str(axes) for axes in axes_vals]
@pytest.mark.parametrize('axes', axes_vals, ids=axes_ids)
def test_some_axes(thr, axes):
    check_norm(thr, (20, 10, 5), numpy.float32, 2, axes)



from logging import error

from mako import exceptions

import reikna.helpers as helpers
from reikna.helpers import template_for, template_from, template_def, extract_signature_and_value
from reikna.cluda import dtypes


TEMPLATE = template_for(__file__)


def render_prelude(thr, fast_math=False):
    return TEMPLATE.get_def('prelude').render(
        api=thr.api.get_id(), compile_fast_math=fast_math)


def render_template(template, *args, **kwds):
    # add some "built-ins" to the kernel
    render_kwds = dict(dtypes=dtypes, helpers=helpers)
    assert set(render_kwds).isdisjoint(set(kwds))
    render_kwds.update(kwds)

    try:
        src = template.render(*args, **render_kwds)
    except:
        error(
            "Failed to render template with"
            "\nargs: {args}\nkwds: {kwds}\nsource:\n{source}\n"
            "{exception}".format(
                args=args, kwds=kwds, source=template.source,
                exception=exceptions.text_error_template().render()))
        raise Exception("Template rendering failed")
    return src


class Snippet:
    """
    Contains a CLUDA snippet.
    See :ref:`tutorial-modules` for details.

    :param template_src: a ``Mako`` template with the module code,
        or a string with the template source.
    :type template_src: ``str`` or ``Mako`` template.
    :param render_kwds: a dictionary which will be used to render the template.
        Can contain other modules and snippets.
    """

    def __init__(self, template_src, render_kwds=None):
        self.template = template_from(template_src)
        self.render_kwds = {} if render_kwds is None else dict(render_kwds)

    @classmethod
    def create(cls, func_or_str, render_kwds=None):
        """
        Creates a snippet from the ``Mako`` def:

        * if ``func_or_str`` is a function, then the def has the same signature as ``func_or_str``,
          and the body equal to the string it returns;
        * if ``func_or_str`` is a string, then the def has empty signature.
        """
        signature, code = extract_signature_and_value(func_or_str)
        return cls(template_def(signature, code), render_kwds=render_kwds)


class Module:
    """
    Contains a CLUDA module.
    See :ref:`tutorial-modules` for details.

    :param template_src: a ``Mako`` template with the module code,
        or a string with the template source.
    :type template_src: ``str`` or ``Mako`` template.
    :param render_kwds: a dictionary which will be used to render the template.
        Can contain other modules and snippets.
    """

    def __init__(self, template_src, render_kwds=None):
        self.template = template_from(template_src)
        self.render_kwds = {} if render_kwds is None else dict(render_kwds)

    @classmethod
    def create(cls, func_or_str, render_kwds=None):
        """
        Creates a module from the ``Mako`` def:

        * if ``func_or_str`` is a function, then the def has the same signature as ``func_or_str``
          (prefix will be passed as the first positional parameter),
          and the body equal to the string it returns;
        * if ``func_or_str`` is a string, then the def has a single positional argument ``prefix``.
          and the body ``code``.
        """
        signature, code = extract_signature_and_value(func_or_str, default_parameters=['prefix'])
        return cls(template_def(signature, code), render_kwds=render_kwds)


class SourceCollector:

    def __init__(self):
        self.constant_modules = {}
        self.sources = []
        self.prefix_counter = 0

    def add_module(self, module_id, tmpl_def, args, render_kwds):

        # This caching serves two purposes.
        # First, it reduces the amount of generated code by not generating
        # the same module several times.
        # Second, if the same module object is used (without arguments) in other modules,
        # the data structures defined in this module will be suitable
        # for functions in these modules.
        if len(args) == 0:
            if module_id in self.constant_modules:
                return self.constant_modules[module_id]

        prefix = "_module" + str(self.prefix_counter) + "_"
        self.prefix_counter += 1

        src = render_template(tmpl_def, prefix, *args, **render_kwds)
        self.sources.append(src)

        if len(args) == 0:
            self.constant_modules[module_id] = prefix

        return prefix

    def get_source(self):
        return "\n\n".join(self.sources)


class RenderableSnippet:

    def __init__(self, tmpl_def, render_kwds):
        self.template_def = tmpl_def
        self.render_kwds = render_kwds

    def __call__(self, *args):
        return render_template(self.template_def, *args, **self.render_kwds)

    def __str__(self):
        return self()


class RenderableModule:

    def __init__(self, collector, module_id, tmpl_def, render_kwds):
        self.module_id = module_id
        self.collector = collector
        self.template_def = tmpl_def
        self.render_kwds = render_kwds

    def __call__(self, *args):
        return self.collector.add_module(
            self.module_id, self.template_def, args, self.render_kwds)

    def __str__(self):
        return self()


def process(obj, collector):
    if isinstance(obj, Snippet):
        render_kwds = process(obj.render_kwds, collector)
        return RenderableSnippet(obj.template, render_kwds)
    elif isinstance(obj, Module):
        render_kwds = process(obj.render_kwds, collector)
        return RenderableModule(collector, id(obj), obj.template, render_kwds)
    elif hasattr(obj, '__process_modules__'):
        return obj.__process_modules__(lambda x: process(x, collector))
    elif isinstance(obj, dict):
        return dict(((k,process(v, collector)) for k, v in obj.items()))
    elif isinstance(obj, tuple):
        return tuple(process(v, collector) for v in obj)
    elif isinstance(obj, list):
        return [process(v, collector) for v in obj]
    else:
        return obj


def render_template_source(src, render_args=None, render_kwds=None):

    if render_args is None:
        render_args = []
    if render_kwds is None:
        render_kwds = {}

    collector = SourceCollector()
    render_args = process(render_args, collector)
    main_renderable = process(Snippet(src, render_kwds=render_kwds), collector)

    main_src = main_renderable(*render_args)

    return collector.get_source() + "\n\n" + main_src



from pytest_threadgen import create_thread_in_tuple, \
    parametrize_thread_tuple, pair_thread_with_doubles, get_threads, get_apis

pytest_plugins = ['pytest_returnvalues']


def pytest_addoption(parser):
    parser.addoption("--api", action="store",
        help="API: cuda/ocl/supported",
        # can't get API list from CLUDA, because if we import it here,
        # it messes up with coverage results
        # (modules get imported before coverage collector starts)
        default="supported", choices=["cuda", "ocl", "supported"])
    parser.addoption("--double", action="store",
        help="Use doubles: no/yes/supported",
        default="supported", choices=["no", "yes", "supported"])
    parser.addoption("--fast-math", dest="fast_math", action="store",
        help="Use fast math (where applicable): no/yes/both",
        default="no", choices=["no", "yes", "both"])
    parser.addoption("--device-include-mask", action="append",
        help="Run tests on matching devices only",
        default=[])
    parser.addoption("--device-exclude-mask", action="append",
        help="Run tests on matching devices only",
        default=[])
    parser.addoption("--platform-include-mask", action="append",
        help="Run tests on matching platforms only",
        default=[])
    parser.addoption("--platform-exclude-mask", action="append",
        help="Run tests on matching platforms only",
        default=[])
    parser.addoption("--include-duplicate-devices", action="store_true",
        help="Run tests on all available devices and not only on uniquely named ones",
        default=False)


pytest_funcarg__thr_and_double = create_thread_in_tuple
pytest_funcarg__thr = create_thread_in_tuple
pytest_funcarg__some_thr = create_thread_in_tuple


def pytest_report_header(config):
    tps, tp_ids = get_threads(config)
    devices = dict((tp.device_id, tp.device_full_name) for tp in tps)
    if len(devices) == 0:
        raise ValueError("No devices match the criteria")

    print("Running tests on:")
    for device_id in sorted(devices):
        print("  " + device_id +  ": " + devices[device_id])


def pytest_generate_tests(metafunc):

    if 'fast_math' in metafunc.funcargnames:
        fm = metafunc.config.option.fast_math
        fms = dict(both=[False, True], no=[False], yes=[True])[fm]
        fm_ids = [{False:'nofm', True:'fm'}[fm] for fm in fms]
        metafunc.parametrize('fast_math', fms, ids=fm_ids)

    if 'thr_and_double' in metafunc.funcargnames:
        parametrize_thread_tuple(metafunc, 'thr_and_double', pair_thread_with_doubles)

    if 'thr' in metafunc.funcargnames:
        tps, tp_ids = get_threads(metafunc.config)
        metafunc.parametrize('thr', tps, ids=tp_ids, indirect=True)

    if 'some_thr' in metafunc.funcargnames:
        # Just some thread for tests that only check thread-independent stuff.
        tps, tp_ids = get_threads(metafunc.config)
        metafunc.parametrize('some_thr', [tps[0]], ids=[tp_ids[0]], indirect=True)

    if 'cluda_api' in metafunc.funcargnames:
        apis, api_ids = get_apis(metafunc.config)
        metafunc.parametrize('cluda_api', apis, ids=api_ids)



import weakref

from reikna.helpers import template_def
from reikna.cluda import Snippet
from reikna.core.signature import Signature, Type, Parameter, Annotation
from reikna.core.transformation_modules import leaf_name, node_connector, module_transformation, \
    module_leaf_macro, module_same_indices, module_combined, kernel_declaration, index_cnames


class TransformationParameter(Type):
    """
    Bases: :py:class:`~reikna.core.Type`

    Represents a typed transformation parameter.
    Can be used as a substitute of an array for functions
    which are only interested in array metadata.
    """

    def __init__(self, trf, name, type_):
        Type.__init__(self, type_.dtype, shape=type_.shape, strides=type_.strides)
        self._trf = weakref.ref(trf)
        self._name = name

    def belongs_to(self, trf):
        return self._trf() is trf

    def __str__(self):
        return self._name


class Transformation:
    """
    A class containing a pure parallel transformation of arrays.
    Some restrictions apply:

    * it cannot use local memory;
    * it cannot use global/local id getters (and depends only on externally passed indices);
    * it cannot have ``'io'`` arguments;
    * it has at least one argument that uses
      :py:attr:`~reikna.core.transformation.KernelParameter.load_same` or
      :py:attr:`~reikna.core.transformation.KernelParameter.store_same`, and does it only once.

    :param parameters: a list of :py:class:`~reikna.core.Parameter` objects.
    :param code: a source template for the transformation.
        Will be wrapped in a template def with positional arguments with the names of
        objects in ``parameters``.
    :param render_kwds: a dictionary with render keywords that will be passed to the snippet.
    :param connectors: a list of parameter names suitable for connection.
        Defaults to all non-scalar parameters.
    """
    def __init__(self, parameters, code, render_kwds=None, connectors=None):

        for param in parameters:
            if param.annotation.input and param.annotation.output:
                raise ValueError(
                    "Transformation cannot have 'io' parameters ('" + param.name + "')")

        self.signature = Signature(parameters)

        for param in self.signature.parameters.values():
            setattr(
                self, param.name,
                TransformationParameter(self, param.name, param.annotation.type))

        if connectors is not None:
            self.connectors = connectors
        else:
            self.connectors = [param.name for param in parameters if param.annotation.array]

        tr_param_names = ['idxs'] + [param.name for param in self.signature.parameters.values()]
        self.snippet = Snippet(template_def(tr_param_names, code), render_kwds=render_kwds)


class Node:
    """
    Represents a node in the transformation tree.
    """

    def __init__(self, input_ntr=None, output_ntr=None):
        self.input_ntr = input_ntr
        self.output_ntr = output_ntr

    def connect(self, ntr):
        if ntr.output:
            return Node(input_ntr=self.input_ntr, output_ntr=ntr)
        else:
            return Node(input_ntr=ntr, output_ntr=self.output_ntr)

    def get_child_names(self):
        get_names = lambda ntr: ([], []) if ntr is None else ntr.get_child_names()
        output_before, output_after = get_names(self.output_ntr)
        input_before, input_after = get_names(self.input_ntr)

        if self.output_ntr is None and self.input_ntr is None:
            return [], []
        elif self.output_ntr is None:
            return input_before, input_after
        elif self.input_ntr is None:
            return output_before, output_after
        else:
            return output_before + output_after, input_before + input_after

    def get_connections(self):
        get_conn = lambda ntr: [] if ntr is None else [ntr]
        return get_conn(self.output_ntr) + get_conn(self.input_ntr)


class NodeTransformation:
    """
    Represents a transformation between two nodes
    (an edge of the transformation tree).
    """

    def __init__(self, connector_node_name, trf, node_from_tr):
        self.trf = trf
        self.connector_node_name = connector_node_name
        self.node_from_tr = dict(node_from_tr)

        # Check that all transformation parameter names are represented,
        # and each of them is only mentioned once.
        tr_names_given = set(self.node_from_tr)
        tr_names_available = set(trf.signature.parameters)
        if tr_names_given != tr_names_available:
            raise ValueError(
                "Supplied transformation names {given} "
                " do not fully coincide with the existing ones: {available}".format(
                    given=tr_names_given, available=tr_names_available))

        # If node names were not guaranteed to be different
        # (which is enforced by Computation.connect()),
        # we would have to make the same check for them as well.

        tr_connectors = [
            tr_name for tr_name, node_name in node_from_tr.items()
            if node_name == connector_node_name]
        # There should be only one transformation parameter corresponding to the connector node
        # Not sure under which circumstances it is not true...
        assert len(tr_connectors) == 1
        tr_connector = tr_connectors[0]

        # Taking into account that 'io' parameters are not allowed for transformations.
        self.output = trf.signature.parameters[tr_connector].annotation.input

    def get_child_names(self):
        # Walking the tree conserving the order of parameters in the transformation.
        node_names = [
            self.node_from_tr[tr_param.name]
            for tr_param in self.trf.signature.parameters.values()]
        connector_idx = node_names.index(self.connector_node_name)
        return node_names[:connector_idx], node_names[connector_idx+1:]

    def translate_node_names(self, translator):
        connector_node_name = translator(self.connector_node_name)
        node_from_tr = dict(
            (tr_name, translator(node_name))
            for tr_name, node_name in self.node_from_tr.items())
        return NodeTransformation(connector_node_name, self.trf, node_from_tr)


class TransformationTree:
    """
    A directed acyclic graph of transformations of the kernel arguments.
    """

    def __init__(self, root_parameters):
        # Preserve order of initial root parameters.
        # These can repeat.
        self.root_names = []

        # Keeping whole parameters, because we want to preserve the default values (if any).
        self.root_parameters = {}

        self.nodes = {} # all nodes of the tree
        self.leaf_parameters = {} # nodes available for connection

        for param in root_parameters:
            self.root_names.append(param.name)
            if param.name in self.root_parameters and param != self.root_parameters[param.name]:
                # Could be an 'io' parameter used for separate 'i' and 'o' parameters
                # in a nested computation.
                # Need to check types and merge.

                new_ann = param.annotation
                old_param = self.root_parameters[param.name]
                old_ann = old_param.annotation

                # FIXME: Not sure when these can be raised
                assert old_ann.type == new_ann.type
                assert old_param.default == param.default

                # Given the old_param != param, the only possible combinations of roles are
                # 'i' and 'o', 'i' and 'io', 'o' and 'io'.
                # In all cases the resulting role is 'io'.
                new_param = Parameter(
                    param.name, Annotation(new_ann.type, 'io'), default=param.default)
                self.root_parameters[param.name] = new_param
                self.leaf_parameters[param.name] = new_param
            else:
                self.nodes[param.name] = Node()
                self.root_parameters[param.name] = param
                self.leaf_parameters[param.name] = param

    def _get_subtree_names(self, names, ignore, visited, leaves_only=False):
        """Helper method for traversing the tree."""

        result = []

        for i, name in enumerate(names):
            if name in ignore or name in visited:
                continue

            visited.add(name)
            ignore_in_children = names[i+1:]

            node = self.nodes[name]
            children_before, children_after = node.get_child_names()

            # If the node has a connection yet is still in the leaf parameters
            # (which means it is an i/o node),
            # we want to preserve its position in the list of parameters
            # of the connected transformation.
            # (It helps to preserve the order of the transformation signature
            # when PureParallel computation is created out of it, among other things)
            # This means that we have two subtrees: for the parameters before the connection,
            # and for the parameters after it.

            subtree_before = self._get_subtree_names(
                children_before, ignore_in_children, visited, leaves_only=leaves_only)
            subtree_after = self._get_subtree_names(
                children_after, ignore_in_children, visited, leaves_only=leaves_only)

            if not leaves_only:
                result.append(name)
            result += subtree_before
            if leaves_only and name in self.leaf_parameters:
                result.append(name)
            result += subtree_after

        return result

    def get_subtree_names(self, root_names=None, leaves_only=False):
        if root_names is None:
            root_names = self.root_names
        return self._get_subtree_names(root_names, [], set(), leaves_only=leaves_only)

    def get_root_annotations(self):
        return dict((name, param.annotation) for name, param in self.root_parameters.items())

    def get_root_parameters(self):
        return [self.root_parameters[name] for name in self.root_names]

    def get_leaf_parameters(self, root_names=None):
        leaf_names = self.get_subtree_names(root_names=root_names, leaves_only=True)
        return [self.leaf_parameters[name] for name in leaf_names]

    def _connect(self, ntr):

        # At this point we assume that ``ntr`` describes a valid connection.
        # All sanity checks are performed in ``connect()``.

        for tr_param in ntr.trf.signature.parameters.values():
            node_name = ntr.node_from_tr[tr_param.name]

            if node_name == ntr.connector_node_name:
                ann = self.leaf_parameters[node_name].annotation
                if ann.input and ann.output:
                # splitting the 'io' leaf
                    updated_role = 'i' if ntr.output else 'o'

                    # Since it is an array parameter, we do not need to worry
                    # about preserving the default value (it can't have one).
                    self.leaf_parameters[node_name] = Parameter(
                        node_name, Annotation(ann.type, role=updated_role))
                else:
                # 'i' or 'o' leaf is hidden by the transformation
                    del self.leaf_parameters[node_name]

            else:
                if (node_name in self.leaf_parameters and
                        self.leaf_parameters[node_name].annotation.array):
                    ann = self.leaf_parameters[node_name].annotation
                    if (ann.input and ntr.output) or (ann.output and not ntr.output):
                    # Joining 'i' and 'o' paths into an 'io' leaf.
                    # Since it is an array parameter, we do not need to worry
                    # about preserving the default value (it can't have one).
                        self.leaf_parameters[node_name] = Parameter(
                            node_name, Annotation(ann.type, role='io'))
                else:
                    self.leaf_parameters[node_name] = tr_param.rename(node_name)

            if node_name not in self.nodes:
                self.nodes[node_name] = Node()

        self.nodes[ntr.connector_node_name] = self.nodes[ntr.connector_node_name].connect(ntr)

    def connect(self, comp_connector, trf, comp_from_tr):

        ntr = NodeTransformation(comp_connector, trf, comp_from_tr)

        # Check that the types of connections are correct
        for tr_name, node_name in comp_from_tr.items():
            if node_name not in self.leaf_parameters:
                if node_name == comp_connector:
                    raise ValueError(
                        "Parameter '" + node_name + "' is not a part of the signature")
                elif node_name in self.nodes:
                    raise ValueError(
                        "Parameter '" + node_name + "' is hidden by transformations")

            if node_name not in self.leaf_parameters:
                # If node names could repeat, we would have to check that
                # transformation parameters with incompatible types are not pointing
                # at the same new node.
                continue

            node_ann = self.leaf_parameters[node_name].annotation
            tr_ann = trf.signature.parameters[tr_name].annotation

            if tr_ann.type != node_ann.type:
                raise ValueError(
                    "Incompatible types of the transformation parameter '{tr_name}' ({tr_type}) "
                    "and the node '{node_name}' ({node_type})".format(
                        node_name=node_name, tr_name=tr_name,
                        node_type=node_ann.type, tr_type=tr_ann.type))

            # No more to check in the case of scalars
            if not tr_ann.array:
                continue

            if node_name == comp_connector:
                if ntr.output and not node_ann.output:
                    raise ValueError("'" + node_name + "' is not an output node")
                if not ntr.output and not node_ann.input:
                    raise ValueError("'" + node_name + "' is not an input node")
            else:
                if ntr.output and node_ann.output:
                    raise ValueError(
                        "Cannot connect transformation parameter '{tr_name}' "
                        "to an existing output node '{node_name}'".format(
                            tr_name=tr_name, node_name=node_name))

        self._connect(ntr)

    def reconnect(self, other_tree, translator=None):
        for ntr in other_tree.connections():
            if translator is not None:
                ntr = ntr.translate_node_names(translator)

            if ntr.connector_node_name not in self.leaf_parameters:
                continue

            # In the nested tree this particular node may only use one data path
            # (input or output), despite it being 'io' in the parent tree.
            # Thus we only need to reconnect the transformation if such data path exists.
            ann = self.leaf_parameters[ntr.connector_node_name].annotation
            if (ntr.output and ann.output) or (not ntr.output and ann.input):
                self._connect(ntr)

    def connections(self):
        node_names = self.get_subtree_names(leaves_only=False)
        for name in node_names:
            node = self.nodes[name]
            for ntr in node.get_connections():
                yield ntr

    def translate(self, translator):
        root_params = self.get_root_parameters()
        new_root_params = [param.rename(translator(param.name)) for param in root_params]
        new_tree = TransformationTree(new_root_params)
        new_tree.reconnect(self, translator=translator)
        return new_tree

    def get_subtree(self, parameters):
        # Unless the user was not messing with undocumented fields,
        # same names will correspond to the same parameters.
        # But if they do not, we better catch it here.
        subtree_params = []
        for param in parameters:
            if param.name in self.root_parameters:
                assert self.root_parameters[param.name].annotation == param.annotation
                # Not using the parameter that came from the user,
                # because we want to preserve the default value that is saved in our parameter.
                subtree_params.append(self.root_parameters[param.name])
            else:
                subtree_params.append(param)

        new_tree = TransformationTree(subtree_params)
        new_tree.reconnect(self)
        return new_tree

    def get_kernel_declaration(self, kernel_name):
        leaf_params = self.get_leaf_parameters()

        decl = kernel_declaration(kernel_name, leaf_params)
        leaf_names = [param.name for param in leaf_params]

        return decl, leaf_names

    def _get_transformation_module(self, annotation, ntr):

        param = Parameter(ntr.connector_node_name, annotation)

        tr_args = [Indices(param.annotation.type.shape)]
        connection_names = []
        for tr_param in ntr.trf.signature.parameters.values():
            connection_name = ntr.node_from_tr[tr_param.name]
            connection_names.append(connection_name)

            if connection_name == ntr.connector_node_name:
                if ntr.output:
                    load_same = node_connector(ntr.output)
                    tr_args.append(KernelParameter(
                        param.name, param.annotation.type, load_same=load_same))
                else:
                    store_same = node_connector(ntr.output)
                    tr_args.append(KernelParameter(
                        param.name, param.annotation.type, store_same=store_same))
            else:
                tr_args.append(self._get_kernel_argobject(connection_name, tr_param.annotation))

        subtree_params = self.get_leaf_parameters([ntr.connector_node_name])

        return module_transformation(ntr.output, param, subtree_params, ntr.trf.snippet, tr_args)

    def _get_connection_modules(self, output, name, annotation):

        node = self.nodes[name]
        param = Parameter(name, annotation)
        ntr = node.output_ntr if output else node.input_ntr

        m_idx = None
        m_same = None
        m_combined = None

        if ntr is None:
            m_idx = module_leaf_macro(output, param)
        else:
            m_idx = self._get_transformation_module(annotation, ntr)

        subtree_params = self.get_leaf_parameters([name])

        # FIXME: this module won't work at the base level (that is, not in a trnsformation)
        # unless 'idx' variables were defined.
        # This behavior was enabled for PureParallel.from_trf(), which defines these variables.
        m_same = module_same_indices(output, param, subtree_params, m_idx)

        m_combined = module_combined(output, param, subtree_params, m_idx)

        return m_idx, m_same, m_combined

    def _get_kernel_argobject(self, name, annotation):
        # Returns a parameter object, which can be passed to the main kernel.

        if not annotation.array:
            return KernelParameter(name, annotation.type)

        load_idx, load_same, load_combined_idx = self._get_connection_modules(
            False, name, annotation)
        store_idx, store_same, store_combined_idx = self._get_connection_modules(
            True, name, annotation)

        return KernelParameter(
            name, annotation.type,
            load_idx=load_idx,
            store_idx=store_idx,
            load_same=load_same,
            store_same=store_same,
            load_combined_idx=load_combined_idx,
            store_combined_idx=store_combined_idx)

    def get_kernel_argobjects(self):
        return [
            self._get_kernel_argobject(name, self.root_parameters[name].annotation)
            for name in self.root_names]


class Indices:
    """
    Encapsulates the information about index variables available for the snippet.
    """

    def __init__(self, shape):
        """__init__()""" # hide the signature from Sphinx
        self._names = index_cnames(shape)

    def __getitem__(self, dim):
        """
        Returns the name of the index varibale for the dimension ``dim``.
        """
        return self._names[dim]

    def all(self):
        """
        Returns the comma-separated list of all index variable names
        (useful for passing the guiding indices verbatim in a load or store call).
        """
        return ', '.join(self._names)


class KernelParameter:
    """
    Providing an interface for accessing kernel arguments in a template.
    Depending on the parameter type, and whether it is used
    inside a computation or a transformation template,
    can have different load/store attributes available.

    .. py:attribute:: name

        Parameter name

    .. py:attribute:: shape
    .. py:attribute:: dtype
    .. py:attribute:: ctype
    .. py:attribute:: strides

        Same as in :py:class:`~reikna.core.Type`.

    .. py:method:: __str__()

        Returns the C kernel parameter name corresponding to this parameter.
        It is the only method available for scalar parameters.

    .. py:attribute:: load_idx

        A module providing a macro with the signature ``(idx0, idx1, ...)``,
        returning the corresponding element of the array.

    .. py:attribute:: store_idx

        A module providing a macro with the signature ``(idx0, idx1, ..., val)``,
        saving ``val`` into the specified position.

    .. py:method:: load_combined_idx(slices)

        A module providing a macro with the signature ``(cidx0, cidx1, ...)``,
        returning the element of the array corresponding to the new slicing of indices
        (e.g. an array with shape ``(2, 3, 4, 5, 6)`` sliced as ``slices=(2, 2, 1)``
        is indexed as an array with shape ``(6, 20, 6)``).

    .. py:method:: store_combined_idx(slices)

        A module providing a macro with the signature ``(cidx0, cidx1, ..., val)``,
        saving ``val`` into the specified position
        corresponding to the new slicing of indices.

    .. py:attribute:: load_same

        A module providing a macro that returns the element of the array
        corresponding to the indices used by the caller of the transformation.

    .. py:attribute:: store_same

        A module providing a macro with the signature ``(val)`` that stores ``val``
        using the indices used by the caller of the transformation.
    """

    def __init__(self, name, type_, load_idx=None, store_idx=None, load_same=None, store_same=None,
            load_combined_idx=None, store_combined_idx=None):
        """__init__()""" # hide the signature from Sphinx

        self._type = type_

        self.shape = type_.shape
        self.strides = type_.strides
        self.dtype = type_.dtype
        self.ctype = type_.ctype

        self._leaf_name = leaf_name(name)
        self.name = name

        if load_idx is not None:
            self.load_idx = load_idx
        if store_idx is not None:
            self.store_idx = store_idx
        if load_same is not None:
            self.load_same = load_same
        if store_same is not None:
            self.store_same = store_same
        if load_combined_idx is not None:
            self.load_combined_idx = load_combined_idx
        if store_combined_idx is not None:
            self.store_combined_idx = store_combined_idx

    def __process_modules__(self, process):
        kwds = {}
        attrs = (
            'load_idx', 'store_idx', 'load_same', 'store_same',
            'load_combined_idx', 'store_combined_idx')
        for attr in attrs:
            if hasattr(self, attr):
                kwds[attr] = process(getattr(self, attr))

        return KernelParameter(self.name, process(self._type), **kwds)

    def __repr__(self):
        attrs = dict(
            load_idx='li', store_idx='si', load_same='ls', store_same='ss',
            load_combined_idx='lci', store_combined_idx='sci')

        attr_str = ", ".join([abbr for name, abbr in attrs.items() if hasattr(self, name)])
        if len(attr_str) > 0:
            attr_str = ", " + attr_str

        return "KernelParameter("+ self.name + attr_str + ")"

    def __str__(self):
        return self._leaf_name



import itertools

import numpy
import pytest

from helpers import *

from reikna.algorithms import Transpose


def pytest_generate_tests(metafunc):
    if 'shape_and_axes' in metafunc.funcargnames:
        dims = [2, 3, 4]
        shape_sets = {
            2: [(13, 11), (511, 134), (77, 2049)],
            3: [(13, 33, 1029), (77, 55, 33)],
            4: [(11, 13, 19, 31), (35, 4, 57, 8)]}
        size_limit = 2 ** 20

        axes = [None, (1, 0), (1, 0, 2), (2, 0, 1), (2, 3, 0, 1), (0, 3, 1, 2)]

        vals = []
        for dim in dims:
            shapes = shape_sets[dim]
            axes_vals = [a for a in axes if a is None or len(a) == dim]
            vals += itertools.product(shapes, axes_vals)

        metafunc.parametrize('shape_and_axes', vals,
            ids=[str(shape) + "," + str(axes) for shape, axes in vals])

def test_errors(thr, shape_and_axes):
    shape, axes = shape_and_axes
    a = get_test_array(shape, numpy.int32)
    a_dev = thr.to_device(a)
    res_ref = numpy.transpose(a, axes)

    tr = Transpose(a, axes=axes)
    res_dev = thr.empty_like(tr.parameter.output)

    tr = tr.compile(thr)
    tr(res_dev, a_dev)

    assert diff_is_negligible(res_dev.get(), res_ref)
